@@ -1,6 +1,5 @@
 # Base folder support
-# Copyright (C) 2002 John Goerzen
-# <jgoerzen@complete.org>
+# Copyright (C) 2002-2015 John Goerzen & contributors
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -16,12 +15,17 @@
 #    along with this program; if not, write to the Free Software
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
-from threading import *
-from IMAP import IMAPFolder
+from sys import exc_info
+from threading import Lock
+from offlineimap import OfflineImapError
+from .IMAP import IMAPFolder
 import os.path
 
-class MappingFolderMixIn:
-    """Helper class to map between Folder() instances where both side assign a uid
+class MappedIMAPFolder(IMAPFolder):
+    """IMAP class to map between Folder() instances where both side assign a uid
+
+    This Folder is used on the local side, while the remote side should
+    be an IMAPFolder.
 
     Instance variables (self.):
       r2l: dict mapping message uids: self.r2l[remoteuid]=localuid
@@ -29,15 +33,18 @@ class MappingFolderMixIn:
       #TODO: what is the difference, how are they used?
       diskr2l: dict mapping message uids: self.r2l[remoteuid]=localuid
       diskl2r: dict mapping message uids: self.r2l[localuid]=remoteuid"""
-    def _initmapping(self):
+
+    def __init__(self, *args, **kwargs):
+        IMAPFolder.__init__(self, *args, **kwargs)
         self.maplock = Lock()
         (self.diskr2l, self.diskl2r) = self._loadmaps()
-        self._mb = self.__class__.__bases__[1]
+        self._mb = IMAPFolder(*args, **kwargs)
+        """Representing the local IMAP Folder using local UIDs"""
 
     def _getmapfilename(self):
         return os.path.join(self.repository.getmapdir(),
                             self.getfolderbasename())
-        
+
     def _loadmaps(self):
         self.maplock.acquire()
         try:
@@ -51,7 +58,11 @@ class MappingFolderMixIn:
                 line = file.readline()
                 if not len(line):
                     break
-                line = line.strip()
+                try:
+                    line = line.strip()
+                except ValueError:
+                    raise Exception("Corrupt line '%s' in UID mapping file '%s'"%
+                        (line, mapfilename)), None, exc_info()[2]
                 (str1, str2) = line.split(':')
                 loc = long(str1)
                 rem = long(str2)
@@ -67,18 +78,26 @@ class MappingFolderMixIn:
         try:
             file = open(mapfilename + ".tmp", 'wt')
             for (key, value) in self.diskl2r.iteritems():
-                file.write("%d:%d\n" % (key, value))
+                file.write("%d:%d\n"% (key, value))
             file.close()
             os.rename(mapfilename + '.tmp', mapfilename)
         finally:
             if dolock: self.maplock.release()
 
     def _uidlist(self, mapping, items):
-        return [mapping[x] for x in items]
+        try:
+            return [mapping[x] for x in items]
+        except KeyError as e:
+            raise OfflineImapError("Could not find UID for msg '{0}' (f:'{1}'."
+                " This is usually a bad thing and should be reported on the ma"
+                "iling list.".format(e.args[0], self),
+                OfflineImapError.ERROR.MESSAGE), None, exc_info()[2]
 
-    def cachemessagelist(self):
-        self._mb.cachemessagelist(self)
-        reallist = self._mb.getmessagelist(self)
+    # Interface from BaseFolder
+    def cachemessagelist(self, min_date=None, min_uid=None):
+        self._mb.cachemessagelist(min_date=min_date, min_uid=min_uid)
+        reallist = self._mb.getmessagelist()
+        self.messagelist = self._mb.messagelist
 
         self.maplock.acquire()
         try:
@@ -86,7 +105,7 @@ class MappingFolderMixIn:
             # summary that have been deleted from the folder.
 
             for luid in self.diskl2r.keys():
-                if not reallist.has_key(luid):
+                if not luid in reallist:
                     ruid = self.diskl2r[luid]
                     del self.diskr2l[ruid]
                     del self.diskl2r[luid]
@@ -99,7 +118,7 @@ class MappingFolderMixIn:
             self.l2r = self.diskl2r.copy()
 
             for luid in reallist.keys():
-                if not self.l2r.has_key(luid):
+                if not luid in self.l2r:
                     ruid = nextneg
                     nextneg -= 1
                     self.l2r[luid] = ruid
@@ -107,12 +126,17 @@ class MappingFolderMixIn:
         finally:
             self.maplock.release()
 
+    def dropmessagelistcache(self):
+        self._mb.dropmessagelistcache()
+
+    # Interface from BaseFolder
     def uidexists(self, ruid):
         """Checks if the (remote) UID exists in this Folder"""
         # This implementation overrides the one in BaseFolder, as it is
         # much more efficient for the mapped case.
         return ruid in self.r2l
 
+    # Interface from BaseFolder
     def getmessageuidlist(self):
         """Gets a list of (remote) UIDs.
         You may have to call cachemessagelist() before calling this function!"""
@@ -120,6 +144,7 @@ class MappingFolderMixIn:
         # much more efficient for the mapped case.
         return self.r2l.keys()
 
+    # Interface from BaseFolder
     def getmessagecount(self):
         """Gets the number of messages in this folder.
         You may have to call cachemessagelist() before calling this function!"""
@@ -127,13 +152,14 @@ class MappingFolderMixIn:
         # much more efficient for the mapped case.
         return len(self.r2l)
 
+    # Interface from BaseFolder
     def getmessagelist(self):
         """Gets the current message list. This function's implementation
         is quite expensive for the mapped UID case.  You must call
         cachemessagelist() before calling this function!"""
 
         retval = {}
-        localhash = self._mb.getmessagelist(self)
+        localhash = self._mb.getmessagelist()
         self.maplock.acquire()
         try:
             for key, value in localhash.items():
@@ -152,10 +178,12 @@ class MappingFolderMixIn:
         finally:
             self.maplock.release()
 
+    # Interface from BaseFolder
     def getmessage(self, uid):
         """Returns the content of the specified message."""
-        return self._mb.getmessage(self, self.r2l[uid])
+        return self._mb.getmessage(self.r2l[uid])
 
+    # Interface from BaseFolder
     def savemessage(self, uid, content, flags, rtime):
         """Writes a new message, with the specified uid.
 
@@ -170,20 +198,26 @@ class MappingFolderMixIn:
         If the uid is > 0, the backend should set the uid to this, if it can.
         If it cannot set the uid to that, it will save it anyway.
         It will return the uid assigned in any case.
+
+        See folder/Base for details. Note that savemessage() does not
+        check against dryrun settings, so you need to ensure that
+        savemessage is never called in a dryrun mode.
         """
+        self.ui.savemessage('imap', uid, flags, self)
         # Mapped UID instances require the source to already have a
         # positive UID, so simply return here.
         if uid < 0:
             return uid
 
-        #if msg uid already exists, just modify the flags
+        # If msg uid already exists, just modify the flags.
         if uid in self.r2l:
             self.savemessageflags(uid, flags)
             return uid
 
-        newluid = self._mb.savemessage(self, -1, content, flags, rtime)
+        newluid = self._mb.savemessage(-1, content, flags, rtime)
         if newluid < 1:
-            raise ValueError("Backend could not find uid for message")
+            raise ValueError("Backend could not find uid for message, "
+                "returned %s"% newluid)
         self.maplock.acquire()
         try:
             self.diskl2r[newluid] = uid
@@ -193,23 +227,59 @@ class MappingFolderMixIn:
             self._savemaps(dolock = 0)
         finally:
             self.maplock.release()
+        return uid
 
+    # Interface from BaseFolder
     def getmessageflags(self, uid):
-        return self._mb.getmessageflags(self, self.r2l[uid])
+        return self._mb.getmessageflags(self.r2l[uid])
 
+    # Interface from BaseFolder
     def getmessagetime(self, uid):
         return None
 
+    # Interface from BaseFolder
     def savemessageflags(self, uid, flags):
-        self._mb.savemessageflags(self, self.r2l[uid], flags)
+        """Note that this function does not check against dryrun settings,
+        so you need to ensure that it is never called in a
+        dryrun mode."""
 
+        self._mb.savemessageflags(self.r2l[uid], flags)
+
+    # Interface from BaseFolder
     def addmessageflags(self, uid, flags):
-        self._mb.addmessageflags(self, self.r2l[uid], flags)
+        self._mb.addmessageflags(self.r2l[uid], flags)
 
+    # Interface from BaseFolder
     def addmessagesflags(self, uidlist, flags):
-        self._mb.addmessagesflags(self, self._uidlist(self.r2l, uidlist),
+        self._mb.addmessagesflags(self._uidlist(self.r2l, uidlist),
                                   flags)
 
+    # Interface from BaseFolder
+    def change_message_uid(self, ruid, new_ruid):
+        """Change the message from existing ruid to new_ruid
+
+        :param new_uid: The old remote UID will be changed to a new
+            UID. The UIDMaps case handles this efficiently by simply
+            changing the mappings file."""
+        if ruid not in self.r2l:
+            raise OfflineImapError("Cannot change unknown Maildir UID %s"%
+                ruid, OfflineImapError.ERROR.MESSAGE)
+        if ruid == new_ruid: return  # sanity check shortcut
+        self.maplock.acquire()
+        try:
+            luid = self.r2l[ruid]
+            self.l2r[luid] = new_ruid
+            del self.r2l[ruid]
+            self.r2l[new_ruid] = luid
+            # TODO: diskl2r|r2l are a pain to sync and should be done away with
+            # diskl2r only contains positive UIDs, so wrap in ifs.
+            if luid > 0: self.diskl2r[luid] = new_ruid
+            if ruid > 0: del self.diskr2l[ruid]
+            if new_ruid > 0: self.diskr2l[new_ruid] = luid
+            self._savemaps(dolock = 0)
+        finally:
+            self.maplock.release()
+
     def _mapped_delete(self, uidlist):
         self.maplock.acquire()
         try:
@@ -227,28 +297,21 @@ class MappingFolderMixIn:
         finally:
             self.maplock.release()
 
+    # Interface from BaseFolder
     def deletemessageflags(self, uid, flags):
-        self._mb.deletemessageflags(self, self.r2l[uid], flags)
+        self._mb.deletemessageflags(self.r2l[uid], flags)
 
+    # Interface from BaseFolder
     def deletemessagesflags(self, uidlist, flags):
-        self._mb.deletemessagesflags(self, self._uidlist(self.r2l, uidlist),
-                                     flags)
+        self._mb.deletemessagesflags(self._uidlist(self.r2l, uidlist),
+            flags)
 
+    # Interface from BaseFolder
     def deletemessage(self, uid):
-        self._mb.deletemessage(self, self.r2l[uid])
+        self._mb.deletemessage(self.r2l[uid])
         self._mapped_delete([uid])
 
+    # Interface from BaseFolder
     def deletemessages(self, uidlist):
-        self._mb.deletemessages(self, self._uidlist(self.r2l, uidlist))
+        self._mb.deletemessages(self._uidlist(self.r2l, uidlist))
         self._mapped_delete(uidlist)
-
-    #def syncmessagesto_neguid_msg(self, uid, dest, applyto, register = 1):
-    # does not need changes because it calls functions that make the changes   
-    # same goes for all other sync messages types.
-    
-
-# Define a class for local part of IMAP.
-class MappedIMAPFolder(MappingFolderMixIn, IMAPFolder):
-    def __init__(self, *args, **kwargs):
-	IMAPFolder.__init__(self, *args, **kwargs)
-        self._initmapping()
